#! /usr/bin/python
"""Program to exercise and score touch typing skills"""

# Written by Craig Van Degrift, craig@yosemitefoothills.com

import termios      # tcgetattr, tcsetattr, ICANON,
                    #   ECHO, TCSANOW, TCSAFLUSH
import fcntl        # fcntl, F_GETFL, F_SETFL
import os           # O_NONBLOCK, system, popen
import sys          # exit, stdin, stdout
import select       # select
import random       # choice 
import time         # time, strftime
import Image        # new
import ImageDraw    # ImageDraw
import ImageFont    # truetype

textColors={'black':'\033[0;30m',\
              'red':'\033[0;31m',\
            'green':'\033[0;32m',\
            'brown':'\033[0;33m',\
             'blue':'\033[0;34m',\
          'magenta':'\033[0;35m',\
             'cyan':'\033[0;36m',\
             'gray':'\033[0;37m',\
      'brightBlack':'\033[1;30m',\
        'brightRed':'\033[1;31m',\
      'brightGreen':'\033[1;32m',\
           'yellow':'\033[1;33m',\
       'brightBlue':'\033[1;34m',\
    'brightMagenta':'\033[1;35m',\
       'brightCyan':'\033[1;36m',\
            'white':'\033[1;37m'}

fileDir='/home/pi/TypingTest/'

def getStudentName():
    initials=raw_input('\x1bc'+'\n'*(screenHeight/20)+textColors['brightGreen']+\
        ' '*(screenWidth/2-12)+'Enter your initials: '+textColors['white'])
    if initials == '':
        print '\n'+textColors['brightRed']+\
                ' '*(screenWidth/2-17)+'No initials!  Exiting program...'
        sys.exit()
    try:
        with open(fileDir+initials+'.scores','r') as f:
            line=f.readline()
            while line:
                lastLine=line
                line=f.readline()
            name=lastLine[0:25].strip()
            defaultLevel=int(lastLine[76:78])
            defaultTimeLimit=int(lastLine[106:111])

    except IOError:
        name=raw_input('\nEnter your full name: ')
        defaultLevel=1
        defaultTimeLimit=30
#    print os.system('feh qwerty.png &')
    return [name,initials,defaultLevel,defaultTimeLimit]
        
def saveScore(score):
    try:
        with open(fileDir+initials+'.scores','a') as f:
            f.write(score)
    except:
        raise

def showTopScores(currentLevel, n=20):
    wpm=[[] for i in range(0,10)]
    indent=' '*((screenWidth-55)/2)
    topScores=dict()
    totalTime=0
    try:
        with open(fileDir+initials+'.scores','r') as f:
            line=f.readline()
            if line:
                name=line[0:25].strip()
                print '\n'+indent+textColors['white']+\
                        'Top scores for '+\
                        '{0} at level {1}'.format(name,currentLevel)
                print indent+textColors['brightGreen']+'    Date and Time      WPM'+\
                        '  Errors Fixes Duration Total'

            while line:
                if name!=line[0:25].strip():
                    print textColors['brightRed']+\
                          '         Different names in score file.\n'\
                          '         Must be fixed!.  Exiting...\n'
                    sys.exit()
                timeString=line[29:50]
                level=int(line[76:78])
                if currentLevel == level:
                    wpm=line[51:54].strip()+line[55]
                    duration=line[82:90].strip()
                    timeLimit=line[106:111]
                    errors=line[118:122].strip()
                    if len(line) > 123:
                        fixes=line[134:138].strip()
                    else:
                        fixes=''
                    totalTime=str(int(totalTime)+int(duration))
                    s = indent+textColors['white']+timeString+\
                            '{0:>6}'.format(wpm)+\
                            '   {0:>3}'.format(errors)
                    if fixes:
                        s += ' {0:>6}'.format(fixes)
                    else:
                        s += '       '
                    s += '    {0:>4}'.format(duration)+\
                            ' {0:>6}'.format(totalTime)
                    lastKey=int(wpm)
                    topScores[lastKey]=s
                line=f.readline()
            topScores[lastKey]=topScores[lastKey].replace(textColors['white'],textColors['yellow'])
            wpmValues=topScores.keys()
            wpmValues.sort()
            wpmValues.reverse()
            for i in wpmValues[0:n]:
                print topScores[i].replace(' '+str(i),str(i)[:-1]+'.'+str(i)[-1])
    except:
        raise

def getAllScores(printing=False):
    plotData=[[] for i in range(0,16)]
    wpm=[[] for i in range(0,16)]
    durations=[[] for i in range(0,16)]
    totalTimes=[[] for i in range(0,16)]
    highestTotalTime=0
    indent=' '*((screenWidth-55)/2)
    try:
        with open(fileDir+initials+'.scores','r') as f:
            line=f.readline()
            if line:
                name=line[0:25].strip()
                if printing:
                    print '\n'+indent+textColors['white']+' '*((46-len(name))/2)+\
                            'Scores for '+'{0}'.format(name)
                    print indent+textColors['yellow']+'    Date and Time      WPM Level'+\
                            ' Errors Fixes Duration Total'

            while line:
                if name!=line[0:25].strip():
                    print textColors['brightRed']+\
                          '         Different names in score file.\n'\
                          '         Must be fixed!.  Exiting...\n'
                    sys.exit()
                timeString=line[29:50]
                level=int(line[76:78])
                wpm[level].append(line[51:56].strip())
                durations[level].append(line[82:90].strip())
                timeLimit=line[106:111]
                errors=line[118:122].strip()
                if len(line) > 123:
                    fixes=line[134:138].strip()
                else:
                    fixes=''
                if totalTimes[level]==[]:
                    totalTimes[level].append(durations[level][-1])
                else:
                    totalTimes[level].\
                        append(str(int(totalTimes[level][-1])+int(durations[level][-1])))
                if int(totalTimes[level][-1]) > highestTotalTime:
                    highestTotalTime=int(totalTimes[level][-1])
                if printing:
                    print indent+textColors['white']+timeString,\
                            '{0:>5}'.format(wpm[level][-1]),\
                            '{0:2d}'.format(level),\
                            '   {0:>3}'.format(errors),
                    if fixes:
                        print ' {0:>6}'.format(fixes),
                    else:
                        print ' '*7,
                    print   '    {0:>4}'.format(durations[level][-1]),\
                            ' {0:>6}\n'.format(totalTimes[level][-1])
                line=f.readline()
            for level in range(1,16):
              plotData[level]=\
                      zip(totalTimes[level],wpm[level],durations[level])
            return plotData+[highestTotalTime]
    except:
        raise

def plotProgress():
#    print 0,
    data=getAllScores(printing=False)
    highestTotalTime=data.pop()/60
    xLabels=[0,10,20,30,40,50,60,70,80,90,100,110]
    if highestTotalTime < 115:
        im = Image.new('RGB',[640,480],'black')
    elif highestTotalTime < 193:
        im = Image.new('RGB',[1024,480],'black')
        xLabels.extend([120,130,140,150,160,170,180,190])
    elif highestTotalTime < 243:
        im = Image.new('RGB',[1270,480],'black')
        xLabels.extend([120,130,140,150,160,170,180,190,200,210,220,230,240])
    else:
        print '\n'+textColors['brightRed']+\
                'Highest total time exceeds graph.  Fix graph routine!'
        sys.exit()
#    print 1,
    draw=ImageDraw.Draw(im)
#    print 2,
    graphCorner=(50,50)
#    print 3,

    labelFont = ImageFont.truetype(fileDir+'DejaVuSans.ttf',18)
    for x in xLabels:
        draw.line([(graphCorner[0]+x*5,10),\
                (graphCorner[0]+x*5,im.size[1]-graphCorner[1])],\
                fill='DimGray')
#        print 4,
        draw.text((graphCorner[0]+x*5-labelFont.getsize(str(x))[0]/2,\
                im.size[1]-graphCorner[1]+5),\
                str(x),\
                fill='white',\
                font=labelFont)
#        print 5,
    xLabel='Total Time (minutes)'
    draw.text(((im.size[0]-labelFont.getsize(xLabel)[0])/2,\
            im.size[1]-25),\
            xLabel,\
            fill='cyan',\
            font=labelFont)
#    print 6,
    yLabels=[0,10,20,30,40,50,60,70,80,90,100]
    for y in yLabels:
        draw.line([(graphCorner[0],im.size[1]-graphCorner[1]-y*4),\
                (im.size[0]-10,im.size[1]-graphCorner[1]-y*4)],\
                fill='DimGray')
#        print 7,
        draw.text(((graphCorner[0]-labelFont.getsize(str(y))[0]-3),\
                im.size[1]-graphCorner[1]-10-y*4),\
                str(y),\
                fill='white',\
                font=labelFont)
#        print 8,
    yLabel='Words Per Minute'
    tempIm=Image.new('RGB', (200,20))
    tempDraw = ImageDraw.Draw(tempIm)
#    print 9,
    tempDraw.text( (0, 0), yLabel,  font=labelFont, fill='cyan')
#    print 10,
    w=tempIm.rotate(90,  expand=1)
    im.paste(w,(0,100))
#    print 11,
    del tempDraw
    del tempIm
    del w

    draw.rectangle([(graphCorner[0],10),\
            (im.size[0]-10,im.size[1]-graphCorner[1])])
#    print 12,
    
    titleFont = ImageFont.truetype(fileDir+'DejaVuSans.ttf',18)
    title='Words Per Minute for '+name
    draw.text((graphCorner[0]+\
            (im.size[0]-76-titleFont.getsize(title)[0]),\
            im.size[1]-76),\
            title,\
            fill='yellow',\
            font=titleFont)
#    print 13,


    pointColors=['','white','yellow','cyan','red','green',\
            'chocolate','gray','violet','skyblue','gold',\
            'limegreen','pink','orange','khaki','lightgreen']
    legendFont = ImageFont.truetype(fileDir+'DejaVuSans.ttf',14)
#    print 14,
    y=im.size[1]-95
    draw.text((im.size[0]-200,y-33),'Legend',fill='white',font=legendFont)
#    print 15,
    for level in range(1,16):
        x=im.size[0]-340+level*20
        draw.rectangle([(x,y),(x+15,y+15)],fill=pointColors[level])
#        print 16,
        draw.text((x,y-16),str(level),fill=pointColors[level],font=legendFont)
#        print 17,
        for i in data[level]:
            points=[]
            points.append((graphCorner[0]+(float(i[0])-float(i[2]))/12,\
                    im.size[1]-graphCorner[1]-float(i[1])*4))
            points.append((graphCorner[0]+float(i[0])/12,\
                    im.size[1]-graphCorner[1]-float(i[1])*4))
            draw.line(points,fill=pointColors[level])
#    print 18,
    im.save(fileDir+initials+'.gif',format=im.format, overwrite=True)
#    print 19,
    os.system('feh '+fileDir+initials+'.gif')
#    print 20,
    del draw
    del im

def getWordList():
    indent=' '*((screenWidth-36)/2)
    print '\n'+textColors['white']+\
            ' '*((screenWidth-len(name)-8)/2)+\
            'Welcome',name
    print '\n'+textColors['brightGreen']+indent+' '*8,\
            'Avalilable Levels\n\n',\
            textColors['brightCyan']+\
          indent,'Level  1: aeiou...        377 words\n',\
          indent,'Level  2: cr  & basic    1023 words\n',\
          indent,'Level  3: fg  & earlier   797 words\n',\
          indent,'Level  4: l   & earlier  1071 words\n',\
          indent,'Level  5: p   & earlier  1081 words\n',\
          indent,'Level  6: m   & earlier  1242 words\n',\
          indent,'Level  7: xb  & earlier  1070 words\n',\
          indent,'Level  8: wv  & earlier  1304 words\n',\
          indent,'Level  9: qjk & earlier   870 words\n',\
          indent,'Level 10: yz  & earlier  1308 words\n',\
          indent,'Level 11: punctuation     235 words\n',\
          indent,'Level 12: miscellaneous  1040 words\n',\
          indent,'Level 13: numbers, etc.\n',\
          indent,'Level 14: test sample     203 words\n',\
          indent,'Level 15: instructions   1115 words\n\n'

    level=0
    while level==0:
        response=raw_input(indent+' '*11+textColors['brightGreen']+\
                'Enter level: '+textColors['white'])
        if response=='':
            level=defaultLevel
            print indent+'         Assuming level {0}'.format(level)
            break
        if response.isdigit() and int(response) in range(1,16):
            level=int(response)
        else:
            print indent+' Level must be a number from 1 to 15.'

    timeLimit=0
    while timeLimit==0:
        response=raw_input(indent+' '*8+textColors['brightGreen']+\
        'Enter duration: '+textColors['white'])
        if response=='':
            timeLimit=defaultTimeLimit
            print indent,'        Assuming duration {0}'.format(timeLimit)
            break
        if response.isdigit() and int(response) in range(30,10001):
            timeLimit=int(response)
        else:
            print indent,'Duration must be from 30 to 10000 seconds.'

    files=['basic','cr','fg','l','p','m','xb','wv','qjk',\
            'yz','punct','misc','numbers','test','inst']
    try:
        with open(fileDir+files[level-1]+'.wl','r') as f:
            if level in range(13):
                wordlist=f.read().strip('\n/').split('/')
            elif 13 <= level <= 15:
                wordlist=f.readlines()
                l=len(wordlist)
                for i in range(l):
                    wordlist[i]=wordlist[i].strip('\r\n')+'\n'
    except:
        raise

    return wordlist+[level]+[timeLimit]

def getNextLine(level):
    longerWordLine=''
    s=''
    if level in range(13):
        while True:
            while len(longerWordLine) < screenWidth-1:
                previousWordLine=longerWordLine
                longerWordLine+=random.choice(wordlist)+' '
            yield previousWordLine+'  '
            longerWordLine=''
        yield longerWordLine
    elif level==13 or level==15:
        for line in wordlist:
            yield line.rstrip(' \r\n')+'   '
    elif level==14:
        while wordlist:
            wordLine = wordlist.pop().rstrip(' \r\n')+' '
            start=0
            while start < len(wordLine):
                end=start+screenWidth - 3
                if end >= len(wordLine):
                    end = len(wordLine)-1
                while wordLine[end] != ' ':
                    end -= 1
                yield wordLine[start:end+1]+'  '
                start=end+1

def getCharacter():
    fd = sys.stdin.fileno()
    oldterm = termios.tcgetattr(fd)
    newattr = oldterm[:]
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)
    oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)
    try:
        while 1:
            r, w, e = select.select([fd], [], [])
            if r:
                c = sys.stdin.read(5)
                return c
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
        fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)

def doTest():
    global duration, count, errors, fixes
    abort=False
    tstart=time.time()
    line=getNextLine(level)
    while duration < timeLimit and not abort:
        typedWord=''
        lastLetterError=False
        try:
            lineOfWords=line.next()
        except(StopIteration):
            line.close()
            break
        print '\n\n',textColors['brightGreen']+lineOfWords+textColors['white']
        if lineOfWords.rstrip(' ')=='':
            print
            continue
        i=0
        typedWord=''
        error=False
        fixed=False
        while i < len(lineOfWords)\
                and duration < timeLimit\
                and not abort:
            c=getCharacter()
            while len(c) !=1:
		if c not in ['\x1b\x4f\x50','\x1b\x4f\x51','\x1b\x4f\x52','\x1b\x4f\x53',\
			 '\x1b\x5b\x31\x35\x7e','\x1b\x5b\x31\x37\x7e','\x1b\x5b\x31\x38\x7e',\
			 '\x1b\x5b\x31\x39\x7e','\x1b\x5b\x32\x30\x7e','\x1b\x5b\x32\x31\x7e',\
			 '\x1b\x5b\x32\x33\x7e','\x1b\x5b\x32\x34\x7e','\x1b\x5b\x36\x3b\x33',\
			 '\x1b\x5b\x32\x7e','\x1b\x5b\x33\x7e','\x1b\x5b\x35\x73','\x1b\x5b\x35\x7e',\
			 '\x1b\x5b\x36\x7e','\x1b\x4f\x46','\x1b\x4f\x48',\
			 '\x1b\x5b\x41','\x1b\x5b\x42','\x1b\x5b\x43','\x1b\x5b\x44',\
			 '\x1b\x5b\x46','\x1b\x5b\x48']:
                    for j in range(len(c)):
                        print ord(c[j]),
			print '{0:3x}'.format(ord(c[j])),
                    print
                c=getCharacter()
            if c == '\n':
               c = ' '
            if c == ' ' and typedWord:
                if error:
                    sys.stdout.write('\x08'*(i-errorStart)+\
                            textColors['brightRed']+\
                            typedWord[errorStart-i:]+
                            textColors['white'])
                    errors+=1
                    error=False
                elif lineOfWords[i] == ' ':
                    count+=len(typedWord)+1
                    typedWord=''
                    sys.stdout.write(' ')
                    duration=(time.time()-tstart)
                    i+=1
                    if fixed:
                        fixes+=1
                    fixed=False
                    if lineOfWords[i:] == ' '*(len(lineOfWords)-i):
                        break
                else:
                    error=True
                    errorStart=i
                    typedWord=typedWord+c
                    sys.stdout.write(c)
                    i+=1
            elif c == '\x1b':
                sys.stdout.flush()
                abort=True
                break
            elif c=='\x7f':
                if len(typedWord) > 0:
                    sys.stdout.write('\x08 \x08')
                    typedWord=typedWord[:-1]
                    i-=1
                    fixed=True
                    lastLetterError=False
                    if error and i==errorStart:
                        error=False
            else:
                if i==len(lineOfWords)-1:
                    if error==False and not lastLetterError:
                        errors+=1
                        fixes+=1
                        lastLetterError=True
                else:
                    if c != lineOfWords[i]:
                        if error==False:
                            error=True
                            errorStart = i
                    typedWord=typedWord+c
                    sys.stdout.write(c)
                    i+=1
            sys.stdout.flush()
    return abort

def doTestCycle():
    global wordlist,timeLimit,level,initials,name,\
            defaultLevel,defaultTimeLimit
    if not initials:
        student=getStudentName()
        name=student[0]
        initials=student[1]
        defaultLevel=student[2]
        defaultTimeLimit=student[3]
    wordlist=getWordList()
    timeLimit=wordlist.pop()
    defaultTimeLimit=timeLimit
    level=wordlist.pop()
    defaultLevel=level
    abort=doTest()
    if duration > 29:
        wpm=60*count/duration/5
    else:
        print textColors['brightRed']+\
                '\n    Escape pressed within 30 seconds, aborting...\n'
        sys.exit()
    indent = ' '*((screenWidth-39)/2)
    if errors==0:
        indent += '  '
    print '\n\n'+' '*((screenWidth-48)/2)+textColors['yellow']+\
            '{0:.1f} words per minute'.format(wpm),

    if errors == 0:
        print textColors['yellow']+'with no errors',
        if fixes == 0:
            print 'and no fixes!'
    elif errors == 1:
        print textColors['brightRed']+'with only 1 error',
    else:
        print textColors['brightRed']+'with {0:d} errors'.format(errors),
    
    if  fixes == 0:
        if errors != 0:
            print textColors['yellow']+'but with no fixes.'
    elif fixes == 1:

        print textColors['brightRed']+'and only 1 fix.'
    else:
        print textColors['brightRed']+'and {0:d} fixes.'.format(fixes)

    print '\n'+textColors['white']+' '*((screenWidth-36)/2),\
          '{0:d} letters typed in {1:.1f} seconds\n'.format(count,duration)

    result=''
    if abort:
        while result != 'discard' and result != 'save':
            result=raw_input(' '*((screenWidth-40)/2)+\
                textColors['brightRed']+'Discard this result? ('+\
                textColors['white']+' discard'+\
                textColors['brightRed']+' or '+
                textColors['white']+'save '+\
                textColors['brightRed']+'): '+textColors['white'])

    if result != 'discard':
        timeString=time.strftime("%b %d, %Y %H:%M:%S")
        score='{0:25}'.format(name)+\
                ' on {0:5}'.format(timeString)+\
                ' {0:5.1f} words/min'.format(wpm)+\
                ' at level {0:2d}'.format(level)+\
                ' for {0:6d} seconds'.format(int(duration))+\
                ' (limit:{0:6d})'.format(timeLimit)+\
                ' with {0:4} errors'.format(errors)+\
                ' and {0:4} fixes\n'.format(fixes)
        saveScore(score)
        if raw_input(' '*((screenWidth-40)/2)+\
                textColors['brightGreen']+'Show your scores and graph? ('+\
                textColors['white']+'Y'+\
                textColors['brightGreen']+' or '+\
                textColors['white']+'n'+\
                textColors['brightGreen']+'): '+\
                textColors['white']) not in ['n','N','no','No','NO','nope']:
            showTopScores(level)
            plotProgress()

rows,columns = os.popen('stty size', 'r').read().split()
screenHeight=int(rows)
screenWidth=int(columns)
initials=''
name=''
wordlist=[]
defaultLevel=0
defaultTimeLimit=0
duration=0
count=0
errors=0
fixes=0
again=True
while again:

#    initials='ctv'
#    name='Craig Van Degrift'
#    level=12
#    showTopScores(level)
#    plotProgress()

    doTestCycle()
    again=raw_input('\n'+textColors['brightGreen']+' '*((screenWidth-20)/2)+\
            'Go again? ('+textColors['white']+'Y'+textColors['brightGreen']+\
            ' or '+textColors['white']+'n'+textColors['brightGreen']+'): '+\
            textColors['white']) not in ['n','N','no','No','NO','nope']
    duration=0
    count=0
    errors=0
    fixes=0



